#Code to reproduce results of manuscript "A New Dimension to Mortality Forecasts" by Benjamin Seligman,
#Gabi Greenberg, and Shripad Tuljapurkar, Demographic Research, 2016

#Li-Lee Coherent Forecasting

# Functions ---------------------------------------------------------------
#Loads HMD files and isolates desired years
HMDLoad<-function(stuff,Yrmin,Yrmax){ #stuff is HMD life table, Yrmin is minimum year, Yrmax is maxmimum year
  if(is.na(Yrmin)==F & stuff$Year[1] < Yrmin){stuff = stuff[which(stuff$Year >= Yrmin),]}
  if(is.na(Yrmax)==F & stuff$Year[nrow(stuff)] > Yrmax){stuff = stuff[which(stuff$Year <= Yrmax),]}
  W<-length(levels(factor(stuff$Age)))
  P<-nrow(stuff)/W
  Q<-vector(mode="numeric")
  for(z in 1:P){
    Q<-cbind(Q,stuff[(z*W-(W-1)):(z*W),3])
  } 
  return(Q)
}

#Calculation of Lee-Carter Model given a matrix of mx's with rows = ages, columns = years
LCcalc<-function(Q){
  logQ<-log(Q)
  A<-rowMeans(logQ)
  X<-logQ-A
  S<-svd(X)
  Vt<-t(S$v)
  Ut<-t(S$u)
  B<-Ut[1,]/sum(Ut[1,])
  #B<-Vt[,1]/sum(Vt[,1]) <-This uses the column space (years), which is not correct
  k<-S$v[,1]*sum(Ut[1,])*S$d[1]
  return(list(a=A,b=B,k=k))
}

#Calculating e0 using rwf forecasts without relying recomposition with SVD
#function to project forward mx given initial mx, series of k's, and b
#Returns forecast k's without launch error using initial k and deltas
launch<-function(init,k){
  series<-c(init,k)
  delts<-tail(series,-1)-head(series,-1)
  return(init+cumsum(delts))
}

#Convert L-C a,b,k back into matrix of mx (rows = age, cols = year)
fctom<-function(a,b,k){
  Q<-as.matrix(b)%*%k
  Q<-Q+matrix(a,ncol=length(k),nrow=length(a))
  return(exp(Q))
}

#Calculate life expectancies at birth (e0) from mx
e0calc<-function(m,sex){
  if(sex=="m"){a0=0.045+2.684*m[1]
  } else if(sex=="f"){a0=0.053+2.8*m[1]
  } else{a0=0.07+1.7*m[1]}
  q1<-m[1]/(1+m[1]*(1-a0))
  q<-m[2:length(m)]/(1+m[2:length(m)]*0.5)
  q<-c(q1,q)
  lx<-1-q
  l=c(1,head(lx,-1))
  l<-cumprod(l)
  ex<-(1-q)*l+q*l*c(a0,rep(0.5,length(q)-1))
  return(sum(ex))
}


# Data Loading and Forecasting --------------------------------------------
#Li Approach: Creates stacked Male/Female mx matrix then does SVD on that for L-C parameters using above function

namevec<-c("CAN","DEUTW","FRATNP","GBR_NP","ITA","JPN","USA")
for(i in 1:length(namevec)){
  Mal<-read.table(paste(namevec[i],".mltper_1x1.txt",sep=""),header=TRUE,skip=2)
  assign(paste(namevec[i],".m.e0",sep=""),cbind(as.numeric(levels(factor(Mal$Year))),as.numeric(Mal$ex[which(Mal$Age==0)])))
  Mobs<-HMDLoad(Mal,1989,NA)
  Mal<-HMDLoad(Mal,Yrmin=1950,Yrmax=1989)
  Fem<-read.table(paste(namevec[i],".fltper_1x1.txt",sep=""),header=TRUE,skip=2)
  assign(paste(namevec[i],".f.e0",sep=""),cbind(as.numeric(levels(factor(Fem$Year))),as.numeric(Fem$ex[which(Fem$Age==0)])))
  Fobs<-HMDLoad(Fem,1989,NA)
  Fem<-HMDLoad(Fem,Yrmin=1950,Yrmax=1989)
  Q<-rbind(Mal,Fem)
  assign(paste(namevec[i],".li",sep=""),LCcalc(Q))
  assign(paste(namevec[i],".liobs",sep=""),rbind(Mobs,Fobs))
}
rm(Mal,Mobs,Fem,Fobs,Q)

#Making 20-year forecasts with Li decompositions
#Forecasting k for the final 20 years of data for each G7
#country in samplew/ 1st stage k's using daya since 1950
library(forecast)
for(i in 1:length(namevec)){
  tlistb<-get(paste(namevec[i],".li",sep=""))[[3]]
  assign(paste(namevec[i],".li.fc",sep=""),rwf(tlistb,h=20,drift=T)) 
}
rm(tlistb)

#Generate observed k's directly & convert forecasts to launch error-corrected forecasts
for(i in 1:length(namevec)){
  a<-get(paste(namevec[i],".li",sep=""))$a
  b<-get(paste(namevec[i],".li",sep=""))$b
  liobs<-log(get(paste(namevec[i],".liobs",sep="")))-a
  liobs<-liobs[,2:ncol(liobs)]
  kobs<-apply(liobs*b,2,sum)/(b%*%b)
  start<-tail(get(paste(namevec[i],".li",sep=""))$k,1)
  kmean<-launch(start,get(paste(namevec[i],".li.fc",sep=""))$mean)
  klow<-launch(start,get(paste(namevec[i],".li.fc",sep=""))$lower[,2])
  khigh<-launch(start,get(paste(namevec[i],".li.fc",sep=""))$upper[,2])
  assign(paste(namevec[i],"klaunch",sep="."),list(o=kobs,m=kmean,l=klow,h=khigh))
  plot(x=seq(1989,2009,1),y=c(start,kmean),type="l",col="red",ylim=c(min(c(kobs,klow)),max(c(kobs,khigh))),xlab="Year",ylab="k",main=paste(namevec[i],"Observed vs Forecast k, Li Method"))
  points(x=seq(1989,1989+length(kobs),1),y=c(start,kobs),pch="+",col="blue")
  lines(x=seq(1989,2009,1),y=c(start,klow),col="red")
  lines(x=seq(1989,2009,1),y=c(start,khigh),col="red")
  delts<-tail(c(start,kobs),-1)-head(c(start,kobs),-1)
  meanie<-get(paste(namevec[i],".li.fc",sep=""))$model$drift
  stdd<-get(paste(namevec[i],".li.fc",sep=""))$model$sd
  chi<-sum(((delts-meanie)/stdd)^2)
  assign(paste(namevec[i],"kchi",sep="."),1-pchisq(chi,df=20))
}
rm(a,b,liobs,kobs,start,kmean,klow,khigh)


# Plotting Results --------------------------------------------------------
#Life Expectancies recovered from forecasts
#Plotting the results for life-expectancy
for(i in 1:length(namevec)){
  k0<-tail(get(paste(namevec[i],"li",sep="."))$k,1)
  kfc<-get(paste(namevec[i],"klaunch",sep="."))
  delts<-list(o=tail(c(k0,kfc$o),-1)-head(c(k0,kfc$o),-1),m=tail(c(k0,kfc$m),-1)-head(c(k0,kfc$m),-1),l=tail(c(k0,kfc$l),-1)-head(c(k0,kfc$l),-1),h=tail(c(k0,kfc$h),-1)-head(c(k0,kfc$h),-1))
  #Males
  b<-get(paste(namevec[i],"li",sep="."))$b[1:111]
  e0true<-get(paste(namevec[i],".m.e0",sep=""))[which(get(paste(namevec[i],".m.e0",sep=""))[,1]>=1989),2]
  start<-e0true[1]
  y0=log(get(paste(namevec[i],"liobs",sep="."))[1:111,1])
  yx<-list(o=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$o)),m=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$m)),l=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$l)),h=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$h)))
  e0mean<-apply(yx$m,2,e0calc,sex="m")
  e0low<-apply(yx$l,2,e0calc,sex="m")
  e0high<-apply(yx$h,2,e0calc,sex="m")
  plot(x=seq(1989,2009,1),y=c(start,e0mean),ylim=c(min(c(e0true,e0high)),max(c(e0true,e0low))),type="l",col="red",xlab="Year",ylab="e0",main=paste(namevec[i],", Males, Observed vs Li L-C Forecast e0",sep=""))
  lines(x=seq(1989,1988+length(e0true)),y=e0true)
  lines(x=seq(1989,2009,1),y=c(start,e0low),col="red")
  lines(x=seq(1989,2009,1),y=c(start,e0high),col="red")
  #Females
  b<-get(paste(namevec[i],"li",sep="."))$b[112:222]
  e0true<-get(paste(namevec[i],".f.e0",sep=""))[which(get(paste(namevec[i],".f.e0",sep=""))[,1]>=1989),2]
  start<-e0true[1]
  y0=log(get(paste(namevec[i],"liobs",sep="."))[112:222,1])
  yx<-list(o=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$o)),m=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$m)),l=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$l)),h=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$h)))
  e0mean<-apply(yx$m,2,e0calc,sex="f")
  e0low<-apply(yx$l,2,e0calc,sex="f")
  e0high<-apply(yx$h,2,e0calc,sex="f")
  plot(x=seq(1989,2009,1),y=c(start,e0mean),ylim=c(min(c(e0true,e0high)),max(c(e0true,e0low))),type="l",col="red",xlab="Year",ylab="e0",main=paste(namevec[i],", Females, Observed vs Li L-C Forecast e0",sep=""))
  lines(x=seq(1989,1988+length(e0true)),y=e0true)
  lines(x=seq(1989,2009,1),y=c(start,e0low),col="red")
  lines(x=seq(1989,2009,1),y=c(start,e0high),col="red")
}