#Code to reproduce results of manuscript "A New Dimension to Mortality Forecasts" by Benjamin Seligman,
#Gabi Greenberg, and Shripad Tuljapurkar, Demographic Research, 2016

#Main Manuscript Results

# Functions ---------------------------------------------------------------
#Key life-table functions
#Loads HMD files and isolates desired years
HMDLoad<-function(stuff,Yrmin,Yrmax){ #stuff is HMD life table, Yrmin is minimum year, Yrmax is maxmimum year
  if(is.na(Yrmin)==F & stuff$Year[1] < Yrmin){stuff = stuff[which(stuff$Year >= Yrmin),]}
  if(is.na(Yrmax)==F & stuff$Year[nrow(stuff)] > Yrmax){stuff = stuff[which(stuff$Year <= Yrmax),]}
  W<-length(levels(factor(stuff$Age)))
  P<-nrow(stuff)/W
  Q<-vector(mode="numeric")
  for(z in 1:P){
    Q<-cbind(Q,stuff[(z*W-(W-1)):(z*W),3])
  } 
  return(Q)
}

#Calculation of Lee-Carter Model given a matrix of mx's with rows = ages, columns = years
LCcalc<-function(Q){
  logQ<-log(Q)
  A<-rowMeans(logQ)
  X<-logQ-A
  S<-svd(X)
  Vt<-t(S$v)
  Ut<-t(S$u)
  B<-Ut[1,]/sum(Ut[1,])
  #B<-Vt[,1]/sum(Vt[,1]) <-This uses the column space (years), which is not correct
  k<-S$v[,1]*sum(Ut[1,])*S$d[1]
  return(list(a=A,b=B,k=k))
}

#Returns forecast k's without launch error using initial k and deltas
launch<-function(init,k){
  series<-c(init,k)
  delts<-tail(series,-1)-head(series,-1)
  return(init+cumsum(delts))
}

#Convert L-C a,b,k back into matrix of mx (rows = age, cols = year)
fctom<-function(a,b,k){
  Q<-as.matrix(b)%*%k
  Q<-Q+matrix(a,ncol=length(k),nrow=length(a))
  return(exp(Q))
}

#Calculate life expectancies at birth (e0) from mx
e0calc<-function(m,sex="b"){
  if(sex=="m"){a0=0.045+2.684*m[1]
  } else if(sex=="f"){a0=0.053+2.8*m[1]
  } else{a0=0.07+1.7*m[1]}
  q1<-m[1]/(1+m[1]*(1-a0))
  q<-m[2:length(m)]/(1+m[2:length(m)]*0.5)
  q<-c(q1,q)
  lx<-1-q
  l=c(1,head(lx,-1))
  l<-cumprod(l)
  ex<-(1-q)*l+q*l*c(a0,rep(0.5,length(q)-1))
  return(sum(ex))
}

#function to calculate life expectancies from a,b,k
LCe0calc<-function(a,b,k,sex){
  m<-exp(b%*%t(k)+a%*%t(rep(1,length(k))))
  if(sex=="m"){q1<-m[1,]/(1+m[1,]*(1-(0.045+2.684*m[1,])))}
  else if(sex=="f"){q1<-m[1,]/(1+m[1,]*(1-(0.053+2.8*m[1,])))}
  else{q1<-m[1,]/(1+m[1,]*(1-(0.07+1.7*m[1,])))}
  q<-m[2:nrow(m),]/(1+m[2:nrow(m),]*0.5)
  q<-rbind(q1,q)
  lx<-1-q
  l=1
  l<-rbind(l,apply(lx,2,cumprod))
  e0<-apply(l,2,sum)
}


# Reading in Data ---------------------------------------------------------
#Loading HMD life tables for G7 - both sexes, males, females - and calculating L-C on each
namevec<-c("CAN","DEUTW","FRATNP","GBR_NP","ITA","JPN","USA")
countries<-c("Canada","West Germany","France","Great Britain","Italy","Japan","USA")
base_start=1950
base_end=1989
for(i in 1:length(namevec)){
  #Both Sexes
  stuff<-read.table(paste(namevec[i],".bltper_1x1.txt",sep=""),header=TRUE,skip=2)
  #Matrix of observed life expectancies from forecast period
  assign(paste(namevec[i],".b.e0",sep=""),cbind(as.numeric(levels(factor(stuff$Year))),as.numeric(stuff$ex[which(stuff$Age==0)])))
  #Matrix of observed age-specific mortality rates to convert to k's
  assign(paste(namevec[i],".b.obs",sep=""),HMDLoad(stuff,base_end,NA))
  #L-C parameters from 1950-1989 mortality rate matrix
  Q<-HMDLoad(stuff,base_start,base_end)
  Years<-as.integer(levels(factor(stuff$Year)))
  Years<-Years[which(Years>=base_start&Years<=base_end)]
  assign(paste(namevec[i],".b",sep=""),LCcalc(Q))
  assign(paste(namevec[i],".b.err",sep=""),apply(HMDLoad(stuff,base_start,base_end)-get(paste(namevec[i],".b",sep=""))$b%*%t(get(paste(namevec[i],".b",sep=""))$k),2,function(x){sqrt(x%*%x)}))
  #Males
  stuff<-read.table(paste(namevec[i],".mltper_1x1.txt",sep=""),header=TRUE,skip=2)
  assign(paste(namevec[i],".m.e0",sep=""),cbind(as.numeric(levels(factor(stuff$Year))),as.numeric(stuff$ex[which(stuff$Age==0)])))
  assign(paste(namevec[i],".m.obs",sep=""),HMDLoad(stuff,base_end,NA))
  Q<-HMDLoad(stuff,base_start,base_end)
  assign(paste(namevec[i],".m",sep=""),LCcalc(Q))
  assign(paste(namevec[i],".m.err",sep=""),apply(HMDLoad(stuff,base_start,base_end)-get(paste(namevec[i],".m",sep=""))$b%*%t(get(paste(namevec[i],".m",sep=""))$k),2,function(x){sqrt(x%*%x)}))
  #Females
  stuff<-read.table(paste(namevec[i],".fltper_1x1.txt",sep=""),header=TRUE,skip=2)
  assign(paste(namevec[i],".f.e0",sep=""),cbind(as.numeric(levels(factor(stuff$Year))),as.numeric(stuff$ex[which(stuff$Age==0)])))
  assign(paste(namevec[i],".f.obs",sep=""),HMDLoad(stuff,base_end,NA))
  Q<-HMDLoad(stuff,base_start,base_end)
  assign(paste(namevec[i],".f",sep=""),LCcalc(Q))
  assign(paste(namevec[i],".f.err",sep=""),apply(HMDLoad(stuff,base_start,base_end)-get(paste(namevec[i],".f",sep=""))$b%*%t(get(paste(namevec[i],".f",sep=""))$k),2,function(x){sqrt(x%*%x)}))
}
rm(Q,stuff,Years)


# Make Forecasts ----------------------------------------------------------
#Making 20-year forecasts with above decompositions
#Forecasting k for the final 20 years of data for each G7
#country in samplew/ 1st stage k's using data since 1950
library(forecast)
for(i in 1:length(namevec)){
  tlistb<-get(paste(namevec[i],".b",sep=""))[[3]]
  tlistf<-get(paste(namevec[i],".f",sep=""))[[3]]
  tlistm<-get(paste(namevec[i],".m",sep=""))[[3]]
  assign(paste(namevec[i],".b.fc",sep=""),rwf(tlistb,h=20,drift=T))
  assign(paste(namevec[i],".f.fc",sep=""),rwf(tlistf,h=20,drift=T))
  assign(paste(namevec[i],".m.fc",sep=""),rwf(tlistm,h=20,drift=T))    
}
rm(i,tlistb,tlistf,tlistm)

#Converting .obs matrices to k's directly
for(i in 1:length(namevec)){
  bobs<-log(get(paste(namevec[i],".b.obs",sep="")))-get(paste(namevec[i],".b",sep=""))$a
  bobs<-bobs[,2:ncol(bobs)] #this is to get rid of the 1989 observations
  bb<-get(paste(namevec[i],".b",sep=""))$b
  mobs<-log(get(paste(namevec[i],".m.obs",sep="")))-get(paste(namevec[i],".m",sep=""))$a
  mobs<-mobs[,2:ncol(mobs)]
  mb<-get(paste(namevec[i],".m",sep=""))$b
  fobs<-log(get(paste(namevec[i],".f.obs",sep="")))-get(paste(namevec[i],".f",sep=""))$a
  fobs<-fobs[,2:ncol(fobs)]
  fb<-get(paste(namevec[i],".f",sep=""))$b
  kobs<-list(b=apply(bobs*bb,2,sum)/(bb%*%bb),m=apply(mobs*mb,2,sum)/(mb%*%mb),f=apply(fobs*fb,2,sum)/(fb%*%fb))
  assign(paste(namevec[i],".kobs",sep=""),list(b=apply(bobs*bb,2,sum)/(bb%*%bb),m=apply(mobs*mb,2,sum)/(mb%*%mb),f=apply(fobs*fb,2,sum)/(fb%*%fb)))
  kb<-cbind(launch(tail(get(paste(namevec[i],".b",sep=""))$k,1),get(paste(namevec[i],".b.fc",sep=""))$upper[,2]),launch(tail(get(paste(namevec[i],".b",sep=""))$k,1),get(paste(namevec[i],".b.fc",sep=""))$mean),launch(tail(get(paste(namevec[i],".b",sep=""))$k,1),get(paste(namevec[i],".b.fc",sep=""))$lower[,2]))
  kf<-cbind(launch(tail(get(paste(namevec[i],".f",sep=""))$k,1),get(paste(namevec[i],".f.fc",sep=""))$upper[,2]),launch(tail(get(paste(namevec[i],".f",sep=""))$k,1),get(paste(namevec[i],".f.fc",sep=""))$mean),launch(tail(get(paste(namevec[i],".f",sep=""))$k,1),get(paste(namevec[i],".f.fc",sep=""))$lower[,2]))
  km<-cbind(launch(tail(get(paste(namevec[i],".m",sep=""))$k,1),get(paste(namevec[i],".m.fc",sep=""))$upper[,2]),launch(tail(get(paste(namevec[i],".m",sep=""))$k,1),get(paste(namevec[i],".m.fc",sep=""))$mean),launch(tail(get(paste(namevec[i],".m",sep=""))$k,1),get(paste(namevec[i],".m.fc",sep=""))$lower[,2]))
  assign(paste(namevec[i],".klaunch",sep=""),list(b=kb,f=kf,m=km))
  startb<-tail(get(paste(namevec[i],".b",sep=""))$k,1)
  startm<-tail(get(paste(namevec[i],".m",sep=""))$k,1)
  startf<-tail(get(paste(namevec[i],".f",sep=""))$k,1)
  deltsb<-tail(c(startb,kobs$b),-1)-head(c(startb,kobs$b),-1)
  deltsm<-tail(c(startm,kobs$m),-1)-head(c(startm,kobs$m),-1)
  deltsf<-tail(c(startf,kobs$f),-1)-head(c(startf,kobs$f),-1)
}
rm(bobs,bb,mobs,mb,fobs,fb,kb,kf,km,kobs,i,startb,startf,startm)

#Recover e0s from forecasts
for(i in 1:length(namevec)){
  for(z in c("b","m","f")){
    a<-get(paste(namevec[i],".",z,sep=""))$a
    b<-get(paste(namevec[i],".",z,sep=""))$b
    if(z=="m"){sex="Males"
    kobs<-get(paste(namevec[i],"kobs",sep="."))$m
    kfc<-get(paste(namevec[i],"klaunch",sep="."))$m}
    else if(z=="f"){sex="Females"
    kobs<-get(paste(namevec[i],"kobs",sep="."))$f
    kfc<-get(paste(namevec[i],"klaunch",sep="."))$f}
    else if(z=="b"){sex="Combined Sexes"
    kobs<-get(paste(namevec[i],"kobs",sep="."))$b
    kfc<-get(paste(namevec[i],"klaunch",sep="."))$b}
    start<-get(paste(namevec[i],z,"e0",sep="."))[which(get(paste(namevec[i],z,"e0",sep="."))[,1]==base_end),2]
    k0<-tail(get(paste(namevec[i],z,sep="."))$k,1)
    y0<-log(get(paste(namevec[i],z,"obs",sep="."))[,1])
    delts<-list(o=tail(c(k0,kobs),-1)-head(c(k0,kobs),-1),m=tail(c(k0,kfc[,2]),-1)-head(c(k0,kfc[,2]),-1),l=tail(c(k0,kfc[,3]),-1)-head(c(k0,kfc[,3]),-1),h=tail(c(k0,kfc[,1]),-1)-head(c(k0,kfc[,1]),-1))
    yx<-list(o=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$o)),m=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$m)),l=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$l)),h=exp(y0+as.matrix(b,ncol=1)%*%cumsum(delts$h)))
    e0obs<-apply(yx$o,2,e0calc,sex=z)
    e0mean<-apply(yx$m,2,e0calc,sex=z)
    e0low<-apply(yx$l,2,e0calc,sex=z)
    e0high<-apply(yx$h,2,e0calc,sex=z)
    assign(paste(namevec[i],z,"e0fc",sep="."),list(o=e0obs,l=e0low,m=e0mean,h=e0high))
  }
}


# Figures from Paper ------------------------------------------------------
#Forecast Plots
for(z in c("b","m","f")){
  #png(paste(z,"fcast.png",sep=""),height=2670,width=1335,res=160)
  pdf(paste(z,"fcast.pdf",sep=""),height=16,width=8)
  par(mfrow=c(4,2))
  for(i in 1:7){
    e0mat=get(paste(namevec[i],z,"e0fc",sep="."))
    e0mat=cbind(e0mat$l,e0mat$m,e0mat$h)
    e0obs<-get(paste(namevec[i],z,"e0",sep="."))
    e0obs=e0obs[which(e0obs[,1]>1988),2]
    if(length(e0obs)<21){e0obs=c(e0obs,rep(NaN,21-length(e0obs)))}
    if(length(e0obs)>21){e0obs=e0obs[1:21]}
    e0mat=rbind(rep(e0obs[1],3),e0mat)
    e0mat=cbind(e0mat,e0obs)
    matplot(x=seq(1989,2009,1),y=e0mat,type="l",lwd=2,lty=c(1,1,1,1),col=c("red","red","red","black"),xlab="Year",ylab="e0",main=countries[i])
  }
  plot.new()
  legend("center",legend=c("Observed e0","Forecast e0"),lwd=c(2,2),lty=c(1,1),col=c("black","red"),cex=2)
  dev.off()
}
rm(z,i,e0mat,e0obs)

#Male Orthogonal Complement
pdf("MError.pdf",height=12,width=8)
par(mfrow=c(4,2))
for(i in 1:length(namevec)){
  k<-get(paste(namevec[i],".m",sep=""))$k
  a<-get(paste(namevec[i],".m",sep=""))$a
  b<-get(paste(namevec[i],".m",sep=""))$b
  k<-k*sqrt(b%*%b)
  err<-get(paste(namevec[i],".m.err",sep=""))*sqrt(b%*%b)
  b<-b/sqrt(b%*%b)
  obs<-log(get(paste(namevec[i],".m.obs",sep="")))-a
  k1<-apply(obs,2,function(x){x%*%b})
  orthog=apply(obs-b%*%t(k1),2,function(x){sqrt(x%*%x)})
  k1=tail(k,1)+cumsum(tail(k1,-1)-head(k1,-1))
  orthog=tail(err,1)+cumsum(tail(orthog,-1)-head(orthog,-1))
  if(i!=2){
    plot(x=seq(1950,1989+length(k1)),y=c(k,k1),col="red",type="l",main=countries[i],ylim=c(min(c(k,k1)),max(c(k,k1))),lwd=2,xlab="Year",ylab="k(t)")
    lines(x=seq(1950,1989+length(k1)),y=c(err,orthog),col="blue",pch="+",lwd=2,type="l")
    abline(v=1989.5,untf=F,lty=3,lwd=1) 
  }
  else if(i==2){
    plot(x=seq(1956,1989+length(k1)),y=c(k,k1),col="red",type="l",main=countries[i],ylim=c(min(c(k,k1)),max(c(k,k1))),lwd=2,xlab="Year",ylab="k(t)")
    lines(x=seq(1956,1989+length(k1)),y=c(err,orthog),col="blue",pch="+",lwd=2,type="l")
    abline(v=1989.5,untf=F,lty=3,lwd=1)
  }
}
plot.new()
legend("center",legend=c("1st Singular Value","Orthogonal Complement"),lwd=c(2,2),col=c("red","blue"),cex=1.3)
dev.off()
